import { useEffect, useState } from 'react'

import useSend from './useSend'

type UseFetchShared = {
    loading: boolean,
    abort?: () => void,
    refetch: () => void,
}

type UseFetchSucced<T> = {
    error: null,
    data: T,
} & UseFetchShared

type UseFetchErrored = {
    error: string,
    data: undefined,
} & UseFetchShared

type UseFetchReturn<T> = UseFetchSucced<T> | UseFetchErrored

const gotError = <T>(res: UseFetchReturn<T>): res is UseFetchErrored => (
    typeof res.error === 'string'
)

const fallbackError = <T>(res: UseFetchReturn<T>) => (
    gotError(res) ? res.error : res.data
)

function useFetch<R, T extends NonNullable<unknown>>(
    url: string,
    method: RequestInit['method'],
    needAuth: boolean,
    guardResponse: (data: unknown) => data is R,
    processResponse: (data: R) => T,
    initialData?: T,
    params?: Omit<RequestInit, 'method'>,
): UseFetchReturn<T> {
    const [data, setData] = useState(initialData)

    const { doSend, loading, error } = useSend(
        url,
        method,
        needAuth,
        guardResponse,
        processResponse,
        true,
        params,
    )

    function refetch() {
        doSend().then(
            data => { if (data !== undefined) setData(data) }
        ).catch( // must never occur
            err => import.meta.env.DEV && console.error('Failed to do fetch request', err)
        )
    }

    useEffect(refetch, [doSend])

    return {
        ...(
            error === null ? ({
                data: data!, error: null,
            }) : ({ data: undefined, error })
        ),
        loading,
        refetch,
    }
}

export type { UseFetchReturn }

export default useFetch

export { gotError, fallbackError }
